import { __awaiter, __decorate } from "tslib";
import { Events, } from '@ag-grid-community/core';
import { dataServerResultFilter, MessageType, normaliseCriteria, toFieldMetadata, } from '@genesislcap/foundation-comms';
import { LifecycleMixin } from '@genesislcap/foundation-utils';
import { attr, customElement, DOM, observable } from '@microsoft/fast-element';
import { gridProGenesisDatasourceEventNames } from '../grid-pro-genesis-datasource';
import { logger } from '../utils';
import { GridProBaseDatasource } from './base.datasource';
import { getServerSideFilterParamsByFieldType } from './server-side.grid-definitions';
import { DataserverServerSideDatasource } from './server-side.resource-dataserver';
import { ReqRepServerSideDatasource } from './server-side.resource-reqrep';
const criteriaDelimiter = ';';
const withoutColumnDefs = null;
const timeoutInMs = 500;
/**
 * A Genesis Datasource element, for server-side | SSRM-compatible data fetching and used exclusively by the GridPro element.
 * @remarks Only supports Server-Side Row Model. Requires `@ag-grid-enterprise/server-side-row-model` setup and valid AG Grid Enterprise license.
 * @alpha
 */
let GridProServerSideDatasource = class GridProServerSideDatasource extends LifecycleMixin(GridProBaseDatasource) {
    constructor() {
        super(...arguments);
        this.applyFuncName = 'applyServerSideTransaction';
        this.applyAsyncFuncName = 'applyServerSideTransactionAsync';
        this.pagination = false;
        /**
         * The starting VIEW_NUMBER of the data to be fetched.
         * @remarks The VIEW_NUMBER isincremented by the grid as the user scrolls or changes pages.
         */
        this.zeroBasedViewNumber = false;
        /**
         * Enable live updates for the grid.
         * @remarks Only works with DATASERVER resources (StreamDatasource) right now.
         */
        this.liveUpdates = false;
        this.request = {};
    }
    resourceNameChanged(oldValue, newValue) {
        if (!oldValue || oldValue === newValue)
            return;
        DOM.queueUpdate(() => __awaiter(this, void 0, void 0, function* () {
            yield this.restart();
        }));
    }
    criteriaChanged(oldCriteria, newCriteria) {
        const criteriaIsNotDuplicate = oldCriteria !== normaliseCriteria(newCriteria, criteriaDelimiter);
        if (this.ssrmDatasource && criteriaIsNotDuplicate) {
            DOM.queueUpdate(() => __awaiter(this, void 0, void 0, function* () {
                yield this.reloadResourceData();
            }));
        }
    }
    connectedCallback() {
        super.connectedCallback();
        const shouldRunConnect = this.shouldRunConnect;
        DOM.queueUpdate(() => {
            if (!shouldRunConnect)
                return;
            this.init();
        });
    }
    disconnectedCallback() {
        super.disconnectedCallback();
        const shouldRunDisconnect = this.shouldRunDisconnect;
        DOM.queueUpdate(() => __awaiter(this, void 0, void 0, function* () {
            if (!shouldRunDisconnect)
                return;
            yield this.destroy();
        }));
    }
    deepClone() {
        const copy = super.deepClone();
        copy.deferredGridOptions = structuredClone(this.deferredGridOptions);
        copy.deferredColumnStates = structuredClone(this.deferredColumnStates);
        return copy;
    }
    /**
     * Resets the grid data while keeping columnDefs and sends a DATA_LOGOFF message if it's a stream.
     * @remarks This is used when the grid is already initialized and we want to reload the data due to a criteria/filter change.
     * @internal
     */
    reloadResourceData(params) {
        var _a, _b, _c, _d, _e, _f;
        return __awaiter(this, void 0, void 0, function* () {
            this.datasource.destroy();
            (_a = this.liveUpdatesStream) === null || _a === void 0 ? void 0 : _a.unsubscribe();
            this.liveUpdatesStream = undefined;
            this.clearRowData(withoutColumnDefs);
            if (params) {
                // this.datasource.stream = undefined;
                const options = Object.assign(Object.assign({}, this.datasourceOptions()), params);
                const initOK = yield this.datasource.init(options, true, false);
                // this.$emit(gridProGenesisDatasourceEventNames.dataInit);
                if (!initOK) {
                    (_b = this.agGrid.gridApi) === null || _b === void 0 ? void 0 : _b.setServerSideDatasource(null);
                    logger.error('Genesis Datasource re-init failed on filtering/sorting.');
                }
            }
            else {
                if (this.ssrmDatasource) {
                    yield this.ssrmDatasource.destroy();
                    this.ssrmDatasource = undefined;
                }
                (_c = this.agGrid.gridApi) === null || _c === void 0 ? void 0 : _c.setServerSideDatasource(null);
                (_d = this.agGrid.gridApi) === null || _d === void 0 ? void 0 : _d.refreshServerSide({ purge: true });
                (_f = (_e = this.agGrid) === null || _e === void 0 ? void 0 : _e.gridApi) === null || _f === void 0 ? void 0 : _f.showLoadingOverlay();
                yield this.init();
            }
        });
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.agGrid) {
                this.agGrid.gridOptions = Object.assign(Object.assign({ getRowId: (params) => {
                        if (!params.data) {
                            return null;
                        }
                        return params.data[this.rowId];
                    }, getRowStyle(params) {
                        var _a, _b;
                        if ((_b = (_a = params.data) === null || _a === void 0 ? void 0 : _a.ROW_REF) === null || _b === void 0 ? void 0 : _b.includes('_deleted')) {
                            return { color: 'grey' };
                        }
                    }, cacheBlockSize: this.maxRows, defaultColDef: {
                        filter: true,
                        resizable: true,
                        sortable: false,
                    }, pagination: this.pagination, paginationPageSize: this.maxRows, rowBuffer: 0, rowModelType: 'serverSide', suppressServerSideInfiniteScroll: false, suppressMultiSort: true, onFilterChanged: (params) => { }, onSortChanged: (params) => { }, onPaginationChanged: (event) => this.onPaginationChanged(event) }, this.agGrid['agGridOptions']), this.deferredGridOptions);
                this.agGrid['initGrid']();
                this.agGrid.addEventListener(Events.EVENT_GRID_READY, () => __awaiter(this, void 0, void 0, function* () {
                    const initOK = yield this.datasource.init(this.datasourceOptions(), true, false);
                    if (!initOK) {
                        logger.debug(`Genesis Datasource init failed for ${this.resourceName}`);
                        this.clearRowData();
                        return;
                    }
                    this.$emit(gridProGenesisDatasourceEventNames.dataInit);
                    this.indexes = this.getResourceIndexes(this.datasource.availableIndexes);
                    const fieldMetadata = toFieldMetadata(this.datasource.originalFieldDef);
                    const agColumnDefs = yield this.getAgColumnDefs(fieldMetadata);
                    this.agGrid.gridApi.setColumnDefs(agColumnDefs);
                    if (this.isRequestServer) {
                        this.ssrmDatasource = new ReqRepServerSideDatasource({
                            createReqRepRequestFunc: this.createReqRepRequest.bind(this),
                            reloadResourceDataFunc: this.reloadResourceData.bind(this),
                            resourceName: this.resourceName,
                            resourceParams: this.params,
                            resourceIndexes: this.indexes,
                            resourceColDefs: this.datasource.originalFieldDef,
                            maxRows: +this.maxRows,
                            maxView: +this.maxView,
                            rowId: this.rowId,
                            pagination: this.pagination,
                            zeroBasedViewNumber: this.zeroBasedViewNumber,
                        });
                    }
                    else {
                        this.ssrmDatasource = new DataserverServerSideDatasource({
                            createDataserverStreamFunc: this.createDataserverStream.bind(this),
                            reloadResourceDataFunc: this.reloadResourceData.bind(this),
                            resourceName: this.resourceName,
                            resourceParams: this.params,
                            resourceIndexes: this.indexes,
                            resourceColDefs: this.datasource.originalFieldDef,
                            maxRows: +this.maxRows,
                            maxView: +this.maxView,
                            rowId: this.rowId,
                            pagination: this.pagination,
                            zeroBasedViewNumber: this.zeroBasedViewNumber,
                        });
                    }
                    this.agGrid.gridApi.setServerSideDatasource(this.ssrmDatasource);
                }), { once: true });
                return;
            }
            logger.warn(`Application not connected, falling back to local columnDefs/rowData`);
        });
    }
    onPaginationChanged(event) {
        var _a;
        if (this.agGrid.gridApi && this.ssrmDatasource) {
            this.ssrmDatasource.isNewPageSize = (_a = event.newPage) !== null && _a !== void 0 ? _a : false;
        }
    }
    destroy() {
        var _a, _b, _c;
        return __awaiter(this, void 0, void 0, function* () {
            this.datasource.destroy();
            (_a = this.liveUpdatesStream) === null || _a === void 0 ? void 0 : _a.unsubscribe();
            this.liveUpdatesStream = undefined;
            this.clearRowData();
            if (this.ssrmDatasource) {
                yield this.ssrmDatasource.destroy();
                this.ssrmDatasource = undefined;
            }
            (_b = this.agGrid.gridApi) === null || _b === void 0 ? void 0 : _b.setServerSideDatasource(null);
            (_c = this.agGrid.gridApi) === null || _c === void 0 ? void 0 : _c.refreshServerSide({ purge: true });
        });
    }
    restart() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.destroy();
            this.init();
        });
    }
    clearRowData(withColumnDefs = true) {
        var _a, _b, _c, _d;
        this.rowData = new Map();
        if (withColumnDefs) {
            (_b = (_a = this.agGrid) === null || _a === void 0 ? void 0 : _a.gridApi) === null || _b === void 0 ? void 0 : _b.setColumnDefs([]);
        }
        this.agTransaction = { remove: [] };
        (_c = this.agGrid.gridApi) === null || _c === void 0 ? void 0 : _c.forEachNode((node) => {
            this.agTransaction.remove.push(node.data);
        });
        (_d = this.agGrid.gridApi) === null || _d === void 0 ? void 0 : _d.applyServerSideTransaction(this.agTransaction);
        this.agTransaction = undefined;
    }
    getResourceIndexes(avaialbleIndexes) {
        const resourceIndexesMap = new Map();
        avaialbleIndexes === null || avaialbleIndexes === void 0 ? void 0 : avaialbleIndexes.forEach((index) => {
            resourceIndexesMap.set(index.NAME, index.FIELDS.split(' '));
        });
        return resourceIndexesMap;
    }
    get params() {
        const params = {
            CRITERIA_MATCH: this.criteria,
            FIELDS: this.fields,
            MAX_ROWS: this.maxRows,
            MAX_VIEW: this.maxView,
            MOVING_VIEW: this.movingView,
            REVERSE: this.reverse,
        };
        const keys = Array.from(this.indexes.keys());
        if (this.orderBy) {
            if (keys.findIndex((key) => key === this.orderBy) > -1) {
                params.ORDER_BY = this.orderBy;
            }
            else {
                const validIndexes = keys.map((key) => key);
                logger.warn('Specified [orderBy] index is not valid and will not be used. See https://learn.genesis.global/docs/database/data-types/index-entities/ - Available indexes:', validIndexes);
            }
        }
        logger.debug('Getting configured params:', params);
        return params;
    }
    get rowModel() {
        return this.agGrid.gridApi.getModel();
    }
    createReqRepRequest(existingParams = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const reqRep = yield this.datasource.snapshot(existingParams);
            return reqRep;
        });
    }
    createDataserverStream(existingParams = null) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.liveUpdates) {
                const streamOnMessage = (message) => __awaiter(this, void 0, void 0, function* () {
                    var _a, _b;
                    const messageType = message.MESSAGE_TYPE;
                    if (messageType === MessageType.QUERY_UPDATE && message.ROW) {
                        if (!this.ssrmDatasource) {
                            return;
                        }
                        const nextMessage = dataServerResultFilter(message);
                        this.agTransaction = { add: [], remove: [], update: [] };
                        this.handleStreamInserts(nextMessage.inserts);
                        this.handleStreamUpdates(nextMessage.updates);
                        this.handleStreamDeletes(nextMessage.deletes);
                        if (message.SEQUENCE_ID > 1) {
                            this.ssrmDatasource.clientRowsCount += (_a = nextMessage.inserts.length) !== null && _a !== void 0 ? _a : 0;
                            this.ssrmDatasource.serverRowsCount += (_b = nextMessage.inserts.length) !== null && _b !== void 0 ? _b : 0;
                        }
                        // this.handleStreamDeletes(nextMessage.deletes);
                        // this.ssrmDatasource.client_ROWS_COUNT -= nextMessage.deletes.length ?? 0;
                        // this.ssrmDatasource.server_ROWS_COUNT -= nextMessage.deletes.length ?? 0;
                        if (this.pagination) {
                            this.applyAllAgTransactions();
                        }
                        else {
                            // TODO: Double-check this bit / DOM.queueUpdate do not work here
                            setTimeout(() => {
                                this.applyAllAgTransactions();
                            }, timeoutInMs);
                        }
                    }
                });
                const onError = (error) => logger.error(error);
                this.liveUpdatesStream = this.connect
                    .stream(this.resourceName, () => { }, onError, existingParams !== null && existingParams !== void 0 ? existingParams : this.params)
                    .subscribe((message) => streamOnMessage(message));
                // return this.connect.stream(
                //   this.resourceName,
                //   (message) => streamOnMessage(message),
                //   onError,
                //   existingParams ?? this.params,
                // );
            }
            yield this.datasource.startStream();
            return this.datasource.stream;
        });
    }
    // private handleStreamDeletesTemp(deletes: RowData[]) {
    //   if (!deletes) return;
    //   deletes.forEach((deleteData) => {
    //     const rowNode = this.agGrid.gridApi.getRowNode(deleteData.ROW_REF);
    //     if (rowNode) {
    //       rowNode.setData({ ...rowNode.data, ROW_REF: deleteData.ROW_REF + '_deleted' });
    //     }
    //   });
    // }
    getAgColumnDefs(fieldsMetadata) {
        return __awaiter(this, void 0, void 0, function* () {
            const colDefsFromGenesisData = this.generateColumnDefsFromMetadata(fieldsMetadata, getServerSideFilterParamsByFieldType);
            colDefsFromGenesisData.map((colDef) => {
                return {
                    sortable: [].concat(...this.indexes.values()).includes(colDef.field),
                };
            });
            const colDefsMergedWithTemplateDefs = yield this.agGrid.mergeAllColumnDefsAndStates(colDefsFromGenesisData, true);
            return colDefsMergedWithTemplateDefs;
        });
    }
    applyTransaction(agTransaction) {
        return this.agGrid.gridApi.applyServerSideTransaction(agTransaction);
    }
    applyTransactionAsync(agTransaction, callback) {
        this.agGrid.gridApi.applyServerSideTransactionAsync(agTransaction, callback);
    }
};
__decorate([
    attr({ mode: 'boolean' })
], GridProServerSideDatasource.prototype, "pagination", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'zero-based-view-number' })
], GridProServerSideDatasource.prototype, "zeroBasedViewNumber", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'live-updates' })
], GridProServerSideDatasource.prototype, "liveUpdates", void 0);
__decorate([
    observable
], GridProServerSideDatasource.prototype, "request", void 0);
GridProServerSideDatasource = __decorate([
    customElement({
        name: 'grid-pro-server-side-datasource',
    })
], GridProServerSideDatasource);
export { GridProServerSideDatasource };
